type GenericTraps = Record<string, any>

type IRequestStrict = {
  route: string
  params: {
    [key: string]: string
  }
  query: {
    [key: string]: string | string[] | undefined
  }
  proxy?: any
} & Request

type IRequest = IRequestStrict & GenericTraps

type RequestHandler<R = IRequest, Args extends Array<any> = any[]> =
  (request: R, ...args: Args) => any

type ResponseHandler<
  ResponseType = Response,
  RequestType = IRequest,
  Args extends any[] = any[]
> = (
  response: ResponseType & any,
  request: RequestType & any,
  ...args: Args
) => any

type StatusErrorObject = {
  error?: string
  [key: string]: any
}

declare class StatusError extends Error {
  status: number;
  [key: string]: any

  constructor(status = 500, body?: StatusErrorObject | string) {
    super(typeof body === 'object' ? body.error : body)
    typeof body === 'object' && Object.assign(this, body)
    this.status = status
  }
}

type ErrorHandler<
  ErrorType extends Error = StatusError,
  RequestType = IRequest, Args extends any[] = any[]
> = (error: ErrorType, request: RequestType, ...args: Args) => any

type RouteEntry = [
  httpMethod: string,
  match: RegExp,
  handlers: RequestHandler[],
  path?: string,
]

type IttyRouterOptions = {
  base?: string
  routes?: RouteEntry[]
} & GenericTraps

type RouterOptions = {
  before?: RequestHandler<any>[]
  catch?: ErrorHandler
  finally?: ResponseHandler[]
} & IttyRouterOptions

type AutoRouterOptions = {
  missing?: RequestHandler
  format?: ResponseHandler
} & RouterOptions

type RequestLike = {
  method: string
  url: string
} & GenericTraps

type Route<
  R = IRequest,
  A extends Array<any> = any[],
> = <
  RequestType = R,
  Args extends Array<any> = A,
>(
  path: string,
  ...handlers: RequestHandler<RequestType, Args>[]
) => IttyRouterType<R, A>

type CustomRoutes<R = Route> = {
  [key: string]: R
}

type IttyRouterType<
  R = IRequest,
  A extends any[] = any[],
  ResponseType = any,
  GlobalRequestType = R,
> = {
  __proto__: IttyRouterType<R>
  routes: RouteEntry[]
  fetch: <Args extends any[] = A>(request: RequestLike, ...extra: Args) => Promise<ResponseType>
  all: Route<GlobalRequestType, A>
  delete: Route<GlobalRequestType, A>
  get: Route<GlobalRequestType, A>
  head: Route<GlobalRequestType, A>
  options: Route<GlobalRequestType, A>
  patch: Route<GlobalRequestType, A>
  post: Route<GlobalRequestType, A>
  put: Route<GlobalRequestType, A>
} & CustomRoutes<Route<GlobalRequestType, A>> & GenericTraps

type RouterType<R = IRequest, Args extends any[] = any[], ResponseType = any> = {
  before?: RequestHandler<any>[]
  catch?: ErrorHandler
  finally?: ResponseHandler[]
} & IttyRouterType<R, Args, ResponseType>

type AutoRouterType<
  R = IRequest,
  Args extends any[] = any[],
  ResponseType = any
> = {
  missing?: RequestHandler
  format?: ResponseHandler
} & RouterType<R, Args, ResponseType>

type HasContent<ContentType> = {
  content: ContentType
} & IRequestStrict

type ResponseFormatter =
  (body?: any, options?: ResponseInit) => Response

interface ErrorLike extends Error {
  status?: number
  [any: string]: any
}

type ErrorBody = string | object

interface ErrorFormatter {
  (statusCode?: number, body?: ErrorBody): Response
  (error: ErrorLike): Response
}

export type { AutoRouterOptions, AutoRouterType, CustomRoutes, ErrorFormatter, ErrorHandler, HasContent, IRequest, IRequestStrict, IttyRouterOptions, IttyRouterType, RequestHandler, RequestLike, ResponseFormatter, ResponseHandler, Route, RouteEntry, RouterOptions, RouterType };
